package main

import (
	"bytes"
	"context"
	"crypto/x509"
	"errors"
	"fmt"
	"net/http"
	"os"
	"os/signal"
	"runtime/debug"
	"strconv"
	"strings"
	"sync"
	"syscall"
	"time"

	"git.frostfs.info/TrueCloudLab/frostfs-http-gw/internal/cache"
	"git.frostfs.info/TrueCloudLab/frostfs-http-gw/internal/data"
	"git.frostfs.info/TrueCloudLab/frostfs-http-gw/internal/handler"
	"git.frostfs.info/TrueCloudLab/frostfs-http-gw/internal/handler/middleware"
	"git.frostfs.info/TrueCloudLab/frostfs-http-gw/internal/logs"
	internalnet "git.frostfs.info/TrueCloudLab/frostfs-http-gw/internal/net"
	"git.frostfs.info/TrueCloudLab/frostfs-http-gw/internal/service/frostfs"
	"git.frostfs.info/TrueCloudLab/frostfs-http-gw/internal/templates"
	"git.frostfs.info/TrueCloudLab/frostfs-http-gw/metrics"
	"git.frostfs.info/TrueCloudLab/frostfs-http-gw/resolver"
	"git.frostfs.info/TrueCloudLab/frostfs-http-gw/tokens"
	"git.frostfs.info/TrueCloudLab/frostfs-http-gw/tree"
	"git.frostfs.info/TrueCloudLab/frostfs-http-gw/utils"
	"git.frostfs.info/TrueCloudLab/frostfs-observability/tracing"
	v2container "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/api/container"
	cid "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/container/id"
	"git.frostfs.info/TrueCloudLab/frostfs-sdk-go/pool"
	treepool "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/pool/tree"
	"git.frostfs.info/TrueCloudLab/frostfs-sdk-go/user"
	"github.com/fasthttp/router"
	"github.com/nspcc-dev/neo-go/cli/flags"
	"github.com/nspcc-dev/neo-go/cli/input"
	"github.com/nspcc-dev/neo-go/pkg/crypto/keys"
	"github.com/nspcc-dev/neo-go/pkg/util"
	"github.com/nspcc-dev/neo-go/pkg/wallet"
	"github.com/panjf2000/ants/v2"
	"github.com/spf13/viper"
	"github.com/valyala/fasthttp"
	"go.opentelemetry.io/otel/trace"
	"go.uber.org/zap"
	"go.uber.org/zap/zapcore"
	"golang.org/x/exp/slices"
)

type (
	app struct {
		ctx            context.Context
		log            *zap.Logger
		pool           *pool.Pool
		treePool       *treepool.Pool
		key            *keys.PrivateKey
		owner          *user.ID
		cfg            *appCfg
		webServer      *fasthttp.Server
		webDone        chan struct{}
		resolver       *resolver.ContainerResolver
		metrics        *gateMetrics
		services       []*metrics.Service
		settings       *appSettings
		loggerSettings *loggerSettings
		bucketCache    *cache.BucketCache
		handle         *handler.Handler
		corsCnrID      cid.ID

		servers       []Server
		unbindServers []ServerInfo
		mu            sync.RWMutex
	}

	loggerSettings struct {
		mu         sync.RWMutex
		appMetrics *metrics.GateMetrics
	}

	// App is an interface for the main gateway function.
	App interface {
		Wait()
		Serve()
	}

	gateMetrics struct {
		logger   *zap.Logger
		provider *metrics.GateMetrics
		mu       sync.RWMutex
		enabled  bool
	}

	// appSettings stores reloading parameters, so it has to provide getters and setters which use RWMutex.
	appSettings struct {
		reconnectInterval time.Duration
		dialerSource      *internalnet.DialerSource
		workerPoolSize    int
		logLevelConfig    *logLevelConfig

		mu                     sync.RWMutex
		defaultTimestamp       bool
		archiveCompression     bool
		clientCut              bool
		returnIndexPage        bool
		indexPageTemplate      string
		bufferMaxSizeForPut    uint64
		namespaceHeader        string
		defaultNamespaces      []string
		cors                   *data.CORSRule
		enableFilepathFallback bool
	}

	tagsConfig struct {
		tagLogs sync.Map
	}

	logLevelConfig struct {
		logLevel   zap.AtomicLevel
		tagsConfig *tagsConfig
	}
)

func newLogLevel(v *viper.Viper) zap.AtomicLevel {
	ll, err := getLogLevel(v)
	if err != nil {
		panic(err.Error())
	}
	atomicLogLevel := zap.NewAtomicLevel()
	atomicLogLevel.SetLevel(ll)
	return atomicLogLevel
}

func newTagsConfig(v *viper.Viper, ll zapcore.Level) *tagsConfig {
	var t tagsConfig
	if err := t.update(v, ll); err != nil {
		// panic here is analogue of the similar panic during common log level initialization.
		panic(err.Error())
	}
	return &t
}

func newLogLevelConfig(lvl zap.AtomicLevel, tagsConfig *tagsConfig) *logLevelConfig {
	return &logLevelConfig{
		logLevel:   lvl,
		tagsConfig: tagsConfig,
	}
}

func (l *logLevelConfig) update(cfg *viper.Viper, log *zap.Logger) {
	if lvl, err := getLogLevel(cfg); err != nil {
		log.Warn(logs.LogLevelWontBeUpdated, zap.Error(err), logs.TagField(logs.TagApp))
	} else {
		l.logLevel.SetLevel(lvl)
	}

	if err := l.tagsConfig.update(cfg, l.logLevel.Level()); err != nil {
		log.Warn(logs.TagsLogConfigWontBeUpdated, zap.Error(err), logs.TagField(logs.TagApp))
	}
}

func (t *tagsConfig) LevelEnabled(tag string, tgtLevel zapcore.Level) bool {
	lvl, ok := t.tagLogs.Load(tag)
	if !ok {
		return false
	}

	return lvl.(zapcore.Level).Enabled(tgtLevel)
}

func (t *tagsConfig) update(cfg *viper.Viper, ll zapcore.Level) error {
	tags, err := fetchLogTagsConfig(cfg, ll)
	if err != nil {
		return err
	}

	t.tagLogs.Range(func(key, value any) bool {
		k := key.(string)
		v := value.(zapcore.Level)

		if lvl, ok := tags[k]; ok {
			if lvl != v {
				t.tagLogs.Store(key, lvl)
			}
		} else {
			t.tagLogs.Delete(key)
			delete(tags, k)
		}
		return true
	})

	for k, v := range tags {
		t.tagLogs.Store(k, v)
	}

	return nil
}

func newApp(ctx context.Context, cfg *appCfg) App {
	logSettings := &loggerSettings{}
	logLevel := newLogLevel(cfg.config())
	tagConfig := newTagsConfig(cfg.config(), logLevel.Level())
	logConfig := newLogLevelConfig(logLevel, tagConfig)
	log := pickLogger(cfg.config(), logConfig.logLevel, logSettings, tagConfig)

	a := &app{
		ctx:            ctx,
		log:            log.logger,
		cfg:            cfg,
		loggerSettings: logSettings,
		webServer:      new(fasthttp.Server),
		webDone:        make(chan struct{}),
		bucketCache:    cache.NewBucketCache(getBucketCacheOptions(cfg.config(), log.logger), cfg.config().GetBool(cfgFeaturesTreePoolNetmapSupport)),
	}

	a.initAppSettings(logConfig)

	// -- setup FastHTTP server --
	a.webServer.Name = "frost-http-gw"
	a.webServer.ReadBufferSize = a.config().GetInt(cfgWebReadBufferSize)
	a.webServer.WriteBufferSize = a.config().GetInt(cfgWebWriteBufferSize)
	a.webServer.ReadTimeout = a.config().GetDuration(cfgWebReadTimeout)
	a.webServer.WriteTimeout = a.config().GetDuration(cfgWebWriteTimeout)
	a.webServer.DisableHeaderNamesNormalizing = true
	a.webServer.NoDefaultServerHeader = true
	a.webServer.NoDefaultContentType = true
	a.webServer.MaxRequestBodySize = a.config().GetInt(cfgWebMaxRequestBodySize)
	a.webServer.DisablePreParseMultipartForm = true
	a.webServer.StreamRequestBody = a.config().GetBool(cfgWebStreamRequestBody)
	// -- -- -- -- -- -- -- -- -- -- -- -- -- --
	a.initPools(ctx)

	var owner user.ID
	user.IDFromKey(&owner, a.key.PrivateKey.PublicKey)
	a.owner = &owner

	a.setRuntimeParameters()

	a.initResolver()
	a.initMetrics()
	a.initTracing(ctx)
	a.initContainers(ctx)

	return a
}

func (a *app) config() *viper.Viper {
	return a.cfg.config()
}

func (a *app) initContainers(ctx context.Context) {
	corsCnrID, err := a.fetchContainerID(ctx, cfgContainersCORS)
	if err != nil {
		a.log.Fatal(logs.CouldNotFetchCORSContainerInfo, zap.Error(err), logs.TagField(logs.TagApp))
	}
	a.corsCnrID = *corsCnrID
}

func (a *app) initAppSettings(lc *logLevelConfig) {
	a.settings = &appSettings{
		reconnectInterval: fetchReconnectInterval(a.config()),
		dialerSource:      getDialerSource(a.log, a.config()),
		workerPoolSize:    a.config().GetInt(cfgWorkerPoolSize),
		logLevelConfig:    lc,
	}
	a.settings.update(a.config(), a.log)
}

func (s *appSettings) update(v *viper.Viper, l *zap.Logger) {
	defaultTimestamp := v.GetBool(cfgUploaderHeaderEnableDefaultTimestamp)
	archiveCompression := fetchArchiveCompression(v)
	returnIndexPage := v.GetBool(cfgIndexPageEnabled)
	clientCut := v.GetBool(cfgClientCut)
	bufferMaxSizeForPut := v.GetUint64(cfgBufferMaxSizeForPut)
	namespaceHeader := v.GetString(cfgResolveNamespaceHeader)
	defaultNamespaces := fetchDefaultNamespaces(v)
	indexPage, indexEnabled := fetchIndexPageTemplate(v, l)
	cors := fetchCORSConfig(v)
	enableFilepathFallback := v.GetBool(cfgFeaturesEnableFilepathFallback)

	s.mu.Lock()
	defer s.mu.Unlock()

	s.defaultTimestamp = defaultTimestamp
	s.archiveCompression = archiveCompression
	s.returnIndexPage = returnIndexPage
	s.clientCut = clientCut
	s.bufferMaxSizeForPut = bufferMaxSizeForPut
	s.namespaceHeader = namespaceHeader
	s.defaultNamespaces = defaultNamespaces
	s.returnIndexPage = indexEnabled
	s.indexPageTemplate = indexPage
	s.cors = cors
	s.enableFilepathFallback = enableFilepathFallback
}

func (s *loggerSettings) DroppedLogsInc() {
	s.mu.RLock()
	defer s.mu.RUnlock()

	if s.appMetrics != nil {
		s.appMetrics.DroppedLogsInc()
	}
}

func (s *loggerSettings) setMetrics(appMetrics *metrics.GateMetrics) {
	s.mu.Lock()
	defer s.mu.Unlock()

	s.appMetrics = appMetrics
}

func (s *appSettings) DefaultTimestamp() bool {
	s.mu.RLock()
	defer s.mu.RUnlock()
	return s.defaultTimestamp
}

func (s *appSettings) ArchiveCompression() bool {
	s.mu.RLock()
	defer s.mu.RUnlock()
	return s.archiveCompression
}

func (s *appSettings) IndexPageEnabled() bool {
	s.mu.RLock()
	defer s.mu.RUnlock()
	return s.returnIndexPage
}

func (s *appSettings) IndexPageTemplate() string {
	s.mu.RLock()
	defer s.mu.RUnlock()
	if s.indexPageTemplate == "" {
		return templates.DefaultIndexTemplate
	}
	return s.indexPageTemplate
}

func (s *appSettings) CORS() *data.CORSRule {
	s.mu.RLock()
	defer s.mu.RUnlock()

	if s.cors == nil {
		return nil
	}

	allowMethods := make([]string, len(s.cors.AllowedMethods))
	copy(allowMethods, s.cors.AllowedMethods)

	allowHeaders := make([]string, len(s.cors.AllowedHeaders))
	copy(allowHeaders, s.cors.AllowedHeaders)

	exposeHeaders := make([]string, len(s.cors.ExposeHeaders))
	copy(exposeHeaders, s.cors.ExposeHeaders)

	allowOrigins := make([]string, len(s.cors.AllowedOrigins))
	copy(allowOrigins, s.cors.AllowedOrigins)

	return &data.CORSRule{
		AllowedOrigins:     allowOrigins,
		AllowedMethods:     allowMethods,
		AllowedHeaders:     allowHeaders,
		ExposeHeaders:      exposeHeaders,
		AllowedCredentials: s.cors.AllowedCredentials,
		MaxAgeSeconds:      s.cors.MaxAgeSeconds,
	}
}

func (s *appSettings) ClientCut() bool {
	s.mu.RLock()
	defer s.mu.RUnlock()
	return s.clientCut
}

func (s *appSettings) BufferMaxSizeForPut() uint64 {
	s.mu.RLock()
	defer s.mu.RUnlock()
	return s.bufferMaxSizeForPut
}

func (s *appSettings) NamespaceHeader() string {
	s.mu.RLock()
	defer s.mu.RUnlock()
	return s.namespaceHeader
}

func (s *appSettings) FormContainerZone(ns string) string {
	s.mu.RLock()
	namespaces := s.defaultNamespaces
	s.mu.RUnlock()
	if slices.Contains(namespaces, ns) {
		return v2container.SysAttributeZoneDefault
	}

	return ns + ".ns"
}

func (s *appSettings) EnableFilepathFallback() bool {
	s.mu.RLock()
	defer s.mu.RUnlock()
	return s.enableFilepathFallback
}

func (a *app) initResolver() {
	var err error
	a.resolver, err = resolver.NewContainerResolver(a.getResolverConfig())
	if err != nil {
		a.log.Fatal(logs.FailedToCreateResolver, zap.Error(err), logs.TagField(logs.TagApp))
	}
}

func (a *app) getResolverConfig() ([]string, *resolver.Config) {
	resolveCfg := &resolver.Config{
		FrostFS:    frostfs.NewResolverFrostFS(a.pool),
		RPCAddress: a.config().GetString(cfgRPCEndpoint),
	}

	order := a.config().GetStringSlice(cfgResolveOrder)
	if resolveCfg.RPCAddress == "" {
		order = remove(order, resolver.NNSResolver)
		a.log.Warn(logs.ResolverNNSWontBeUsedSinceRPCEndpointIsntProvided, logs.TagField(logs.TagApp))
	}

	if len(order) == 0 {
		a.log.Info(logs.ContainerResolverWillBeDisabledBecauseOfResolversResolverOrderIsEmpty,
			logs.TagField(logs.TagApp))
	}

	return order, resolveCfg
}

func (a *app) initMetrics() {
	gateMetricsProvider := metrics.NewGateMetrics(a.pool)
	a.metrics = newGateMetrics(a.log, gateMetricsProvider, a.config().GetBool(cfgPrometheusEnabled))
	a.metrics.SetHealth(metrics.HealthStatusStarting)
	a.loggerSettings.setMetrics(a.metrics.provider)
}

func newGateMetrics(logger *zap.Logger, provider *metrics.GateMetrics, enabled bool) *gateMetrics {
	if !enabled {
		logger.Warn(logs.MetricsAreDisabled, logs.TagField(logs.TagApp))
	}
	return &gateMetrics{
		logger:   logger,
		provider: provider,
		enabled:  enabled,
	}
}

func (m *gateMetrics) isEnabled() bool {
	m.mu.RLock()
	defer m.mu.RUnlock()

	return m.enabled
}

func (m *gateMetrics) SetEnabled(enabled bool) {
	if !enabled {
		m.logger.Warn(logs.MetricsAreDisabled, logs.TagField(logs.TagApp))
	}

	m.mu.Lock()
	m.enabled = enabled
	m.mu.Unlock()
}

func (m *gateMetrics) SetHealth(status metrics.HealthStatus) {
	if !m.isEnabled() {
		return
	}

	m.provider.SetHealth(status)
}

func (m *gateMetrics) SetVersion(ver string) {
	if !m.isEnabled() {
		return
	}

	m.provider.SetVersion(ver)
}

func (m *gateMetrics) Shutdown() {
	m.mu.Lock()
	if m.enabled {
		m.provider.SetHealth(metrics.HealthStatusShuttingDown)
		m.enabled = false
	}
	m.provider.Unregister()
	m.mu.Unlock()
}

func (m *gateMetrics) MarkHealthy(endpoint string) {
	if !m.isEnabled() {
		return
	}

	m.provider.MarkHealthy(endpoint)
}

func (m *gateMetrics) MarkUnhealthy(endpoint string) {
	if !m.isEnabled() {
		return
	}

	m.provider.MarkUnhealthy(endpoint)
}

func remove(list []string, element string) []string {
	for i, item := range list {
		if item == element {
			return append(list[:i], list[i+1:]...)
		}
	}
	return list
}

func getFrostFSKey(cfg *viper.Viper, log *zap.Logger) (*keys.PrivateKey, error) {
	walletPath := cfg.GetString(cfgWalletPath)

	if len(walletPath) == 0 {
		log.Info(logs.NoWalletPathSpecifiedCreatingEphemeralKeyAutomaticallyForThisRun, logs.TagField(logs.TagApp))
		key, err := keys.NewPrivateKey()
		if err != nil {
			return nil, err
		}
		return key, nil
	}
	w, err := wallet.NewWalletFromFile(walletPath)
	if err != nil {
		return nil, err
	}

	var password *string
	if cfg.IsSet(cfgWalletPassphrase) {
		pwd := cfg.GetString(cfgWalletPassphrase)
		password = &pwd
	}

	address := cfg.GetString(cfgWalletAddress)

	return getKeyFromWallet(w, address, password)
}

func getKeyFromWallet(w *wallet.Wallet, addrStr string, password *string) (*keys.PrivateKey, error) {
	var addr util.Uint160
	var err error

	if addrStr == "" {
		addr = w.GetChangeAddress()
	} else {
		addr, err = flags.ParseAddress(addrStr)
		if err != nil {
			return nil, fmt.Errorf("invalid address")
		}
	}

	acc := w.GetAccount(addr)
	if acc == nil {
		return nil, fmt.Errorf("couldn't find wallet account for %s", addrStr)
	}

	if password == nil {
		pwd, err := input.ReadPassword("Enter password > ")
		if err != nil {
			return nil, fmt.Errorf("couldn't read password")
		}
		password = &pwd
	}

	if err := acc.Decrypt(*password, w.Scrypt); err != nil {
		return nil, fmt.Errorf("couldn't decrypt account: %w", err)
	}

	return acc.PrivateKey(), nil
}

func (a *app) Wait() {
	a.log.Info(logs.StartingApplication,
		zap.String("app_name", "frostfs-http-gw"),
		zap.String("version", Version),
		logs.TagField(logs.TagApp))

	a.metrics.SetVersion(Version)
	a.setHealthStatus()

	<-a.webDone // wait for web-server to be stopped
}

func (a *app) setHealthStatus() {
	a.metrics.SetHealth(metrics.HealthStatusReady)
}

func (a *app) Serve() {
	workerPool := a.initWorkerPool()
	defer func() {
		workerPool.Release()
		close(a.webDone)
	}()

	// Configure router.
	a.configureRouter(workerPool)

	a.startServices()
	a.initServers(a.ctx)

	servs := a.getServers()

	for i := range servs {
		go func(i int) {
			a.log.Info(logs.StartingServer, zap.String("address", servs[i].Address()), logs.TagField(logs.TagApp))
			if err := a.webServer.Serve(servs[i].Listener()); err != nil && err != http.ErrServerClosed {
				a.metrics.MarkUnhealthy(servs[i].Address())
				a.log.Fatal(logs.ListenAndServe, zap.Error(err), logs.TagField(logs.TagApp))
			}
		}(i)
	}

	if len(a.unbindServers) != 0 {
		a.scheduleReconnect(a.ctx, a.webServer)
	}

	sigs := make(chan os.Signal, 1)
	signal.Notify(sigs, syscall.SIGHUP)

LOOP:
	for {
		select {
		case <-a.ctx.Done():
			break LOOP
		case <-sigs:
			a.configReload(a.ctx)
		}
	}

	a.log.Info(logs.ShuttingDownWebServer, zap.Error(a.webServer.Shutdown()), logs.TagField(logs.TagApp))

	a.metrics.Shutdown()
	a.stopServices()
	a.shutdownTracing()
}

func (a *app) initWorkerPool() *ants.Pool {
	workerPool, err := ants.NewPool(a.settings.workerPoolSize)
	if err != nil {
		a.log.Fatal(logs.FailedToCreateWorkerPool, zap.Error(err), logs.TagField(logs.TagApp))
	}
	return workerPool
}

func (a *app) shutdownTracing() {
	const tracingShutdownTimeout = 5 * time.Second
	shdnCtx, cancel := context.WithTimeout(context.Background(), tracingShutdownTimeout)
	defer cancel()

	if err := tracing.Shutdown(shdnCtx); err != nil {
		a.log.Warn(logs.FailedToShutdownTracing, zap.Error(err), logs.TagField(logs.TagApp))
	}
}

func (a *app) configReload(ctx context.Context) {
	a.log.Info(logs.SIGHUPConfigReloadStarted, logs.TagField(logs.TagApp))
	if !a.config().IsSet(cmdConfig) && !a.config().IsSet(cmdConfigDir) {
		a.log.Warn(logs.FailedToReloadConfigBecauseItsMissed, logs.TagField(logs.TagApp))
		return
	}
	if err := a.cfg.reload(); err != nil {
		a.log.Warn(logs.FailedToReloadConfig, zap.Error(err), logs.TagField(logs.TagApp))
		return
	}

	a.settings.logLevelConfig.update(a.cfg.settings, a.log)

	if err := a.settings.dialerSource.Update(fetchMultinetConfig(a.config(), a.log)); err != nil {
		a.log.Warn(logs.MultinetConfigWontBeUpdated, zap.Error(err), logs.TagField(logs.TagApp))
	}

	if err := a.resolver.UpdateResolvers(a.getResolverConfig()); err != nil {
		a.log.Warn(logs.FailedToUpdateResolvers, zap.Error(err), logs.TagField(logs.TagApp))
	}

	if err := a.updateServers(); err != nil {
		a.log.Warn(logs.FailedToReloadServerParameters, zap.Error(err), logs.TagField(logs.TagApp))
	}

	a.setRuntimeParameters()

	a.stopServices()
	a.startServices()

	a.settings.update(a.config(), a.log)

	a.metrics.SetEnabled(a.config().GetBool(cfgPrometheusEnabled))
	a.initTracing(ctx)
	a.setHealthStatus()

	a.log.Info(logs.SIGHUPConfigReloadCompleted, logs.TagField(logs.TagApp))
}

func (a *app) startServices() {
	a.services = a.services[:0]

	pprofConfig := metrics.Config{Enabled: a.config().GetBool(cfgPprofEnabled), Address: a.config().GetString(cfgPprofAddress)}
	pprofService := metrics.NewPprofService(a.log, pprofConfig)
	a.services = append(a.services, pprofService)
	go pprofService.Start()

	prometheusConfig := metrics.Config{Enabled: a.config().GetBool(cfgPrometheusEnabled), Address: a.config().GetString(cfgPrometheusAddress)}
	prometheusService := metrics.NewPrometheusService(a.log, prometheusConfig)
	a.services = append(a.services, prometheusService)
	go prometheusService.Start()
}

func (a *app) stopServices() {
	ctx, cancel := context.WithTimeout(context.Background(), defaultShutdownTimeout)
	defer cancel()

	for _, svc := range a.services {
		svc.ShutDown(ctx)
	}
}

func (a *app) configureRouter(workerPool *ants.Pool) {
	a.handle = handler.New(a.AppParams(), a.settings, tree.NewTree(frostfs.NewPoolWrapper(a.treePool)), workerPool)

	r := router.New()
	r.RedirectTrailingSlash = true
	r.NotFound = func(r *fasthttp.RequestCtx) {
		handler.ResponseError(r, "Not found", fasthttp.StatusNotFound)
	}
	r.MethodNotAllowed = func(r *fasthttp.RequestCtx) {
		handler.ResponseError(r, "Method Not Allowed", fasthttp.StatusMethodNotAllowed)
	}

	r.POST("/upload/{cid}", a.addMiddlewares(a.handle.Upload))
	r.OPTIONS("/upload/{cid}", a.addPreflight(a.handle.Preflight))
	a.log.Info(logs.AddedPathUploadCid, logs.TagField(logs.TagApp))
	r.GET("/get/{cid}/{oid:*}", a.addMiddlewares(a.handle.DownloadByAddressOrBucketName))
	r.HEAD("/get/{cid}/{oid:*}", a.addMiddlewares(a.handle.HeadByAddressOrBucketName))
	r.OPTIONS("/get/{cid}/{oid:*}", a.addPreflight(a.handle.Preflight))
	a.log.Info(logs.AddedPathGetCidOid, logs.TagField(logs.TagApp))
	r.GET("/get_by_attribute/{cid}/{attr_key}/{attr_val:*}", a.addMiddlewares(a.handle.DownloadByAttribute))
	r.HEAD("/get_by_attribute/{cid}/{attr_key}/{attr_val:*}", a.addMiddlewares(a.handle.HeadByAttribute))
	r.OPTIONS("/get_by_attribute/{cid}/{attr_key}/{attr_val:*}", a.addPreflight(a.handle.Preflight))
	a.log.Info(logs.AddedPathGetByAttributeCidAttrKeyAttrVal, logs.TagField(logs.TagApp))
	r.GET("/zip/{cid}/{prefix:*}", a.addMiddlewares(a.handle.DownloadZip))
	r.OPTIONS("/zip/{cid}/{prefix:*}", a.addPreflight(a.handle.Preflight))
	r.GET("/tar/{cid}/{prefix:*}", a.addMiddlewares(a.handle.DownloadTar))
	r.OPTIONS("/tar/{cid}/{prefix:*}", a.addPreflight(a.handle.Preflight))
	a.log.Info(logs.AddedPathZipCidPrefix, logs.TagField(logs.TagApp))

	a.webServer.Handler = r.Handler
}

func (a *app) addMiddlewares(h fasthttp.RequestHandler) fasthttp.RequestHandler {
	list := []func(fasthttp.RequestHandler) fasthttp.RequestHandler{
		a.tracer,
		a.logger,
		a.canonicalizer,
		a.tokenizer,
		a.reqNamespace,
		a.cors,
	}

	for i := len(list) - 1; i >= 0; i-- {
		h = list[i](h)
	}

	return h
}

func (a *app) addPreflight(h fasthttp.RequestHandler) fasthttp.RequestHandler {
	list := []func(fasthttp.RequestHandler) fasthttp.RequestHandler{
		a.tracer,
		a.logger,
		a.canonicalizer,
		a.reqNamespace,
	}

	for i := len(list) - 1; i >= 0; i-- {
		h = list[i](h)
	}

	return h
}

func (a *app) cors(h fasthttp.RequestHandler) fasthttp.RequestHandler {
	return func(c *fasthttp.RequestCtx) {
		h(c)
		code := c.Response.StatusCode()
		if code >= fasthttp.StatusOK && code < fasthttp.StatusMultipleChoices {
			a.handle.SetCORSHeaders(c)
		}
	}
}

func (a *app) logger(h fasthttp.RequestHandler) fasthttp.RequestHandler {
	return func(req *fasthttp.RequestCtx) {
		requiredFields := []zap.Field{zap.Uint64("id", req.ID())}
		reqCtx := utils.GetContextFromRequest(req)
		if traceID := trace.SpanFromContext(reqCtx).SpanContext().TraceID(); traceID.IsValid() {
			requiredFields = append(requiredFields, zap.String("trace_id", traceID.String()))
		}
		log := a.log.With(requiredFields...)

		reqCtx = utils.SetReqLog(reqCtx, log)
		utils.SetContextToRequest(reqCtx, req)

		log.Info(logs.Request, zap.String("remote", req.RemoteAddr().String()),
			zap.ByteString("method", req.Method()),
			zap.ByteString("path", req.Path()),
			zap.ByteString("query", req.QueryArgs().QueryString()),
			logs.TagField(logs.TagDatapath))
		h(req)
	}
}

func (a *app) canonicalizer(h fasthttp.RequestHandler) fasthttp.RequestHandler {
	return func(req *fasthttp.RequestCtx) {
		// regardless of DisableHeaderNamesNormalizing setting, some headers
		// MUST be normalized in order to process execution. They are normalized
		// here.

		toAddKeys := make([][]byte, 0, 10)
		toAddValues := make([][]byte, 0, 10)
		prefix := []byte(utils.UserAttributeHeaderPrefix)

		req.Request.Header.VisitAll(func(k, v []byte) {
			if bytes.HasPrefix(k, prefix) {
				return
			}
			toAddKeys = append(toAddKeys, k)
			toAddValues = append(toAddValues, v)
		})

		// this is safe to do after all headers were read into header structure
		req.Request.Header.EnableNormalizing()

		for i := range toAddKeys {
			req.Request.Header.SetBytesKV(toAddKeys[i], toAddValues[i])
		}

		// return normalization setting back
		req.Request.Header.DisableNormalizing()

		h(req)
	}
}

func (a *app) tokenizer(h fasthttp.RequestHandler) fasthttp.RequestHandler {
	return func(req *fasthttp.RequestCtx) {
		reqCtx := utils.GetContextFromRequest(req)
		appCtx, err := tokens.StoreBearerTokenAppCtx(reqCtx, req)
		if err != nil {
			log := utils.GetReqLogOrDefault(reqCtx, a.log)

			log.Error(logs.CouldNotFetchAndStoreBearerToken, zap.Error(err), logs.TagField(logs.TagDatapath))
			handler.ResponseError(req, "could not fetch and store bearer token: "+err.Error(), fasthttp.StatusBadRequest)
			return
		}
		utils.SetContextToRequest(appCtx, req)
		h(req)
	}
}

func (a *app) tracer(h fasthttp.RequestHandler) fasthttp.RequestHandler {
	return func(req *fasthttp.RequestCtx) {
		appCtx, span := utils.StartHTTPServerSpan(a.ctx, req, "REQUEST")
		defer func() {
			utils.SetHTTPTraceInfo(appCtx, span, req)
			span.End()
		}()

		appCtx = treepool.SetRequestID(appCtx, strconv.FormatUint(req.ID(), 10))

		utils.SetContextToRequest(appCtx, req)
		h(req)
	}
}

func (a *app) reqNamespace(h fasthttp.RequestHandler) fasthttp.RequestHandler {
	return func(req *fasthttp.RequestCtx) {
		appCtx := utils.GetContextFromRequest(req)

		nsBytes := req.Request.Header.Peek(a.settings.NamespaceHeader())
		appCtx = middleware.SetNamespace(appCtx, string(nsBytes))

		utils.SetContextToRequest(appCtx, req)
		h(req)
	}
}

func (a *app) AppParams() *handler.AppParams {
	return &handler.AppParams{
		Logger:    a.log,
		FrostFS:   frostfs.NewFrostFS(a.pool),
		Owner:     a.owner,
		Resolver:  a.resolver,
		Cache:     a.bucketCache,
		CORSCnrID: a.corsCnrID,
		CORSCache: cache.NewCORSCache(getCORSCacheOptions(a.config(), a.log)),
	}
}

func (a *app) initServers(ctx context.Context) {
	serversInfo := fetchServers(a.config(), a.log)

	a.servers = make([]Server, 0, len(serversInfo))
	for _, serverInfo := range serversInfo {
		fields := []zap.Field{
			zap.String("address", serverInfo.Address), zap.Bool("tls enabled", serverInfo.TLS.Enabled),
			zap.String("tls cert", serverInfo.TLS.CertFile), zap.String("tls key", serverInfo.TLS.KeyFile),
		}
		srv, err := newServer(ctx, serverInfo)
		if err != nil {
			a.unbindServers = append(a.unbindServers, serverInfo)
			a.metrics.MarkUnhealthy(serverInfo.Address)
			a.log.Warn(logs.FailedToAddServer, append(fields, zap.Error(err), logs.TagField(logs.TagApp))...)
			continue
		}
		a.metrics.MarkHealthy(serverInfo.Address)

		a.servers = append(a.servers, srv)
		a.log.Info(logs.AddServer, append(fields, logs.TagField(logs.TagApp))...)
	}

	if len(a.servers) == 0 {
		a.log.Fatal(logs.NoHealthyServers, logs.TagField(logs.TagApp))
	}
}

func (a *app) updateServers() error {
	serversInfo := fetchServers(a.config(), a.log)

	a.mu.Lock()
	defer a.mu.Unlock()

	var found bool
	for _, serverInfo := range serversInfo {
		ser := a.getServer(serverInfo.Address)
		if ser != nil {
			if serverInfo.TLS.Enabled {
				if err := ser.UpdateCert(serverInfo.TLS.CertFile, serverInfo.TLS.KeyFile); err != nil {
					return fmt.Errorf("failed to update tls certs: %w", err)
				}
			}
			found = true
		} else if unbind := a.updateUnbindServerInfo(serverInfo); unbind {
			found = true
		}
	}

	if !found {
		return fmt.Errorf("invalid servers configuration: no known server found")
	}

	return nil
}

func (a *app) getServers() []Server {
	a.mu.RLock()
	defer a.mu.RUnlock()
	return a.servers
}

func (a *app) getServer(address string) Server {
	for i := range a.servers {
		if a.servers[i].Address() == address {
			return a.servers[i]
		}
	}
	return nil
}

func (a *app) updateUnbindServerInfo(info ServerInfo) bool {
	for i := range a.unbindServers {
		if a.unbindServers[i].Address == info.Address {
			a.unbindServers[i] = info
			return true
		}
	}
	return false
}

func (a *app) initTracing(ctx context.Context) {
	instanceID := ""
	if len(a.servers) > 0 {
		instanceID = a.servers[0].Address()
	}
	cfg := tracing.Config{
		Enabled:    a.config().GetBool(cfgTracingEnabled),
		Exporter:   tracing.Exporter(a.config().GetString(cfgTracingExporter)),
		Endpoint:   a.config().GetString(cfgTracingEndpoint),
		Service:    "frostfs-http-gw",
		InstanceID: instanceID,
		Version:    Version,
	}

	if trustedCa := a.config().GetString(cfgTracingTrustedCa); trustedCa != "" {
		caBytes, err := os.ReadFile(trustedCa)
		if err != nil {
			a.log.Warn(logs.FailedToInitializeTracing, zap.Error(err), logs.TagField(logs.TagApp))
			return
		}
		certPool := x509.NewCertPool()
		ok := certPool.AppendCertsFromPEM(caBytes)
		if !ok {
			a.log.Warn(logs.FailedToInitializeTracing, zap.String("error", "can't fill cert pool by ca cert"),
				logs.TagField(logs.TagApp))
			return
		}
		cfg.ServerCaCertPool = certPool
	}

	attributes, err := fetchTracingAttributes(a.config())
	if err != nil {
		a.log.Warn(logs.FailedToInitializeTracing, zap.Error(err), logs.TagField(logs.TagApp))
		return
	}
	cfg.Attributes = attributes

	updated, err := tracing.Setup(ctx, cfg)
	if err != nil {
		a.log.Warn(logs.FailedToInitializeTracing, zap.Error(err), logs.TagField(logs.TagApp))
	}
	if updated {
		a.log.Info(logs.TracingConfigUpdated, logs.TagField(logs.TagApp))
	}
}

func (a *app) setRuntimeParameters() {
	if len(os.Getenv("GOMEMLIMIT")) != 0 {
		// default limit < yaml limit < app env limit < GOMEMLIMIT
		a.log.Warn(logs.RuntimeSoftMemoryDefinedWithGOMEMLIMIT, logs.TagField(logs.TagApp))
		return
	}

	softMemoryLimit := fetchSoftMemoryLimit(a.config())
	previous := debug.SetMemoryLimit(softMemoryLimit)
	if softMemoryLimit != previous {
		a.log.Info(logs.RuntimeSoftMemoryLimitUpdated,
			zap.Int64("new_value", softMemoryLimit),
			zap.Int64("old_value", previous),
			logs.TagField(logs.TagApp))
	}
}

func (a *app) scheduleReconnect(ctx context.Context, srv *fasthttp.Server) {
	go func() {
		t := time.NewTicker(a.settings.reconnectInterval)
		defer t.Stop()
		for {
			select {
			case <-t.C:
				if a.tryReconnect(ctx, srv) {
					return
				}
				t.Reset(a.settings.reconnectInterval)
			case <-ctx.Done():
				return
			}
		}
	}()
}

func (a *app) tryReconnect(ctx context.Context, sr *fasthttp.Server) bool {
	a.mu.Lock()
	defer a.mu.Unlock()

	a.log.Info(logs.ServerReconnecting, logs.TagField(logs.TagApp))
	var failedServers []ServerInfo

	for _, serverInfo := range a.unbindServers {
		srv, err := newServer(ctx, serverInfo)
		if err != nil {
			a.log.Warn(logs.ServerReconnectFailed, zap.Error(err), logs.TagField(logs.TagApp))
			failedServers = append(failedServers, serverInfo)
			a.metrics.MarkUnhealthy(serverInfo.Address)
			continue
		}

		go func() {
			a.log.Info(logs.StartingServer, zap.String("address", srv.Address()), logs.TagField(logs.TagApp))
			a.metrics.MarkHealthy(serverInfo.Address)
			if err = sr.Serve(srv.Listener()); err != nil && !errors.Is(err, http.ErrServerClosed) {
				a.log.Warn(logs.ListenAndServe, zap.Error(err), logs.TagField(logs.TagApp))
				a.metrics.MarkUnhealthy(serverInfo.Address)
			}
		}()

		a.servers = append(a.servers, srv)
		a.log.Info(logs.ServerReconnectedSuccessfully,
			zap.String("address", serverInfo.Address), zap.Bool("tls enabled", serverInfo.TLS.Enabled),
			zap.String("tls cert", serverInfo.TLS.CertFile), zap.String("tls key", serverInfo.TLS.KeyFile),
			logs.TagField(logs.TagApp))
	}

	a.unbindServers = failedServers

	return len(a.unbindServers) == 0
}

func (a *app) fetchContainerID(ctx context.Context, cfgKey string) (id *cid.ID, err error) {
	cnrID, err := a.resolveContainerID(ctx, cfgKey)
	if err != nil {
		return nil, err
	}

	err = checkContainerExists(ctx, *cnrID, a.pool)
	if err != nil {
		return nil, err
	}

	return cnrID, nil
}

func (a *app) resolveContainerID(ctx context.Context, cfgKey string) (*cid.ID, error) {
	containerString := a.config().GetString(cfgKey)

	id := new(cid.ID)
	if err := id.DecodeString(containerString); err != nil {
		i := strings.Index(containerString, ".")
		if i < 0 {
			return nil, fmt.Errorf("invalid container address: %s", containerString)
		}

		if id, err = a.resolver.Resolve(ctx, containerString[i+1:], containerString[:i]); err != nil {
			return nil, fmt.Errorf("resolve container address %s: %w", containerString, err)
		}
	}

	return id, nil
}

func checkContainerExists(ctx context.Context, id cid.ID, frostFSPool *pool.Pool) error {
	prm := pool.PrmContainerGet{
		ContainerID: id,
	}

	_, err := frostFSPool.GetContainer(ctx, prm)
	return err
}
